package cc.fascinated.common;

import kong.unirest.core.Headers;
import kong.unirest.core.HttpResponse;
import kong.unirest.core.Unirest;
import lombok.extern.log4j.Log4j2;

import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
@Log4j2(topic = "Request")
public class Request {
    /**
     * The rate limit headers.
     */
    private static final List<String> rateLimitHeaders = List.of(
            "X-RateLimit-Remaining",
            "RateLimit-Remaining"
    );

    /**
     * The rate limit reset headers.
     */
    private static final List<String> rateLimitResetHeaders = List.of(
            "X-RateLimit-Reset",
            "RateLimit-Reset"
    );

    /**
     * Sends a GET request to a URL.
     *
     * @param url the URL to send the request to
     * @param clazz the class to parse the response to
     * @param <T> the type of the response
     * @return the response
     */
    public static <T> HttpResponse<T> get(String url, Class<T> clazz) {
        HttpResponse<T> response = Unirest.get(url).asObject(clazz);
        int rateLimitRemaining = getRateLimitRemaining(response);
        if (rateLimitRemaining == 0) {
            long rateLimitReset = getRateLimitReset(response);
            long timeLeft = rateLimitReset - System.currentTimeMillis();
            try {
                Thread.sleep(timeLeft);
            } catch (InterruptedException e) {
                log.error("Failed to sleep for rate limit reset", e);
            }
            response = Unirest.get(url).asObject(clazz);
        }
        return response;
    }

    /**
     * Gets the rate limit remaining.
     *
     * @param response the response to get the rate limit remaining from
     * @return the rate limit remaining
     */
    public static int getRateLimitRemaining(HttpResponse<?> response) {
        Headers headers = response.getHeaders();
        for (String rateLimitHeader : rateLimitHeaders) {
            if (headers.containsKey(rateLimitHeader)) {
                return Integer.parseInt(headers.getFirst(rateLimitHeader));
            }
        }
        return -1;
    }

    /**
     * Gets the rate limit reset absolute time.
     *
     * @param response the response to get the rate limit reset time from
     * @return the rate limit reset time
     */
    public static long getRateLimitReset(HttpResponse<?> response) {
        Headers headers = response.getHeaders();
        for (String rateLimitResetHeader : rateLimitResetHeaders) {
            if (headers.containsKey(rateLimitResetHeader)) {
                long reset = Long.parseLong(headers.getFirst(rateLimitResetHeader));
                if (reset < 86400) {// Assume it's in seconds left
                    return System.currentTimeMillis() + reset * 1000;
                }
                return reset * 1000; // Assume it's in seconds
            }
        }
        return -1;
    }
}
