package cc.fascinated.services;

import cc.fascinated.common.Request;
import cc.fascinated.exception.impl.BadRequestException;
import cc.fascinated.model.token.ScoreSaberAccountToken;
import cc.fascinated.model.token.ScoreSaberLeaderboardPageToken;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.user.User;
import cc.fascinated.repository.mongo.ScoreSaberLeaderboardRepository;
import kong.unirest.core.HttpResponse;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

/**
 * @author Fascinated (fascinated7)
 */
@Service
@Log4j2(topic = "ScoreSaber Service")
public class ScoreSaberService {
    private static final String SCORESABER_API = "https://scoresaber.com/api/";
    private static final String GET_PLAYER_ENDPOINT = SCORESABER_API + "player/%s/full";
    private static final String GET_LEADERBOARD_ENDPOINT = SCORESABER_API + "leaderboard/by-id/%s/info";
    private static final String GET_RANKED_LEADERBOARDS_ENDPOINT = SCORESABER_API + "leaderboards?ranked=true&page=%s&withMetadata=true";

    /**
     * The ScoreSaber leaderboard repository to use.
     */
    @NonNull
    private final ScoreSaberLeaderboardRepository leaderboardRepository;

    @Autowired
    public ScoreSaberService(@NonNull ScoreSaberLeaderboardRepository leaderboardRepository) {
        this.leaderboardRepository = leaderboardRepository;
    }

    /**
     * Gets the account for a user.
     *
     * @param user the user to get the account for
     * @return the ScoreSaber account
     * @throws BadRequestException if an error occurred while getting the account
     */
    public ScoreSaberAccountToken getAccount(User user) {
        if (user.getSteamId() == null) {
            throw new BadRequestException("%s does not have a linked ScoreSaber account".formatted(user.getUsername()));
        }

        HttpResponse<ScoreSaberAccountToken> response = Request.get(GET_PLAYER_ENDPOINT.formatted(user.getSteamId()), ScoreSaberAccountToken.class);
        if (response.getParsingError().isPresent()) { // Failed to parse the response
            throw new BadRequestException("Failed to parse ScoreSaber account for '%s'".formatted(user.getUsername()));
        }
        if (response.getStatus() != 200) { // The response was not successful
            throw new BadRequestException("Failed to get ScoreSaber account for '%s'".formatted(user.getUsername()));
        }
        return response.getBody();
    }

    /**
     * Gets a leaderboard for a leaderboard id.
     *
     * @param leaderboardId the leaderboard id to get the leaderboard for
     * @return the ScoreSaber leaderboard
     * @throws BadRequestException if an error occurred while getting the leaderboard
     */
    public ScoreSaberLeaderboardToken getLeaderboard(String leaderboardId, boolean bypassCache) {
        Optional<ScoreSaberLeaderboardToken> leaderboardOptional = leaderboardRepository.findById(leaderboardId);
        if (leaderboardOptional.isPresent() && !bypassCache) { // The leaderboard is cached
            return leaderboardOptional.get();
        }

        HttpResponse<ScoreSaberLeaderboardToken> response = Request.get(GET_LEADERBOARD_ENDPOINT.formatted(leaderboardId), ScoreSaberLeaderboardToken.class);
        if (response.getParsingError().isPresent()) { // Failed to parse the response
            throw new BadRequestException("Failed to parse ScoreSaber leaderboard for '%s'".formatted(leaderboardId));
        }
        if (response.getStatus() != 200) { // The response was not successful
            throw new BadRequestException("Failed to get ScoreSaber leaderboard for '%s'".formatted(leaderboardId));
        }
        ScoreSaberLeaderboardToken leaderboard = response.getBody();
        leaderboardRepository.save(leaderboard);
        return leaderboard;
    }

    /**
     * Gets a list of all the ranked leaderboards.
     *
     * @return the ranked leaderboards
     * @throws BadRequestException if an error occurred while getting the leaderboards
     */
    public List<ScoreSaberLeaderboardPageToken> getRankedLeaderboards() {
        log.info("Getting all ranked leaderboards...");
        List<ScoreSaberLeaderboardPageToken> pages = new LinkedList<>();
        int page = 1;
        do {
            log.info("Getting ranked leaderboard page '%s'...".formatted(page));
            ScoreSaberLeaderboardPageToken pageToken = getRankedLeaderboards(page);
            pages.add(pageToken);
            for (ScoreSaberLeaderboardToken leaderboard : pageToken.getLeaderboards()) {
                this.leaderboardRepository.save(leaderboard);
            }
            page++;
        } while (page <= ((pages.get(0).getMetadata().getTotal() / pages.get(0).getMetadata().getItemsPerPage()) + 1));
        log.info("Finished getting all ranked leaderboards, found '{}' pages.", pages.size());
        return pages;
    }

    /**
     * Gets a list of the ranked leaderboards for a page.
     *
     * @param page the page to get the leaderboards for
     * @return the ranked leaderboards
     * @throws BadRequestException if an error occurred while getting the leaderboards
     */
    public ScoreSaberLeaderboardPageToken getRankedLeaderboards(int page) {
        HttpResponse<ScoreSaberLeaderboardPageToken> response = Request.get(GET_RANKED_LEADERBOARDS_ENDPOINT.formatted(page), ScoreSaberLeaderboardPageToken.class);
        if (response.getParsingError().isPresent()) { // Failed to parse the response
            throw new BadRequestException("Failed to parse ScoreSaber leaderboard page for page '%s'".formatted(page));
        }
        if (response.getStatus() != 200) { // The response was not successful
            throw new BadRequestException("Failed to get ScoreSaber leaderboard page for page '%s'".formatted(page));
        }
        return response.getBody();
    }

    /**
     * Gets a leaderboard for a leaderboard id.
     *
     * @param leaderboardId the leaderboard id to get the leaderboard for
     * @return the ScoreSaber leaderboard
     * @throws BadRequestException if an error occurred while getting the leaderboard
     */
    public ScoreSaberLeaderboardToken getLeaderboard(String leaderboardId) {
        return getLeaderboard(leaderboardId, false);
    }
}
