package cc.fascinated.websocket.impl;

import cc.fascinated.common.ScoreSaberUtils;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.token.ScoreSaberPlayerScoreToken;
import cc.fascinated.model.token.ScoreSaberScoreToken;
import cc.fascinated.model.token.ScoreSaberWebsocketDataToken;
import cc.fascinated.platform.Platform;
import cc.fascinated.platform.impl.ScoreSaberPlatform;
import cc.fascinated.services.PlatformService;
import cc.fascinated.services.QuestDBService;
import cc.fascinated.services.UserService;
import cc.fascinated.websocket.Websocket;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.questdb.client.Sender;
import lombok.NonNull;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.socket.TextMessage;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2(topic = "ScoreSaber Websocket")
public class ScoreSaberWebsocket extends Websocket {
    /**
     * The Jackson deserializer to use.
     */
    private final ObjectMapper objectMapper;

    /**
     * The user service to use
     */
    private final UserService userService;

    /**
     * The Influx service to use
     */
    private final QuestDBService questDBService;

    /**
     * The platform service to use
     */
    private final PlatformService platformService;

    @Autowired
    public ScoreSaberWebsocket(@NonNull ObjectMapper objectMapper, @NonNull UserService userService, @NonNull QuestDBService questDBService,
                               @NonNull PlatformService platformService) {
        super("ScoreSaber", "wss://scoresaber.com/ws");
        this.objectMapper = objectMapper;
        this.userService = userService;
        this.questDBService = questDBService;
        this.platformService = platformService;
    }

    @Override @SneakyThrows
    public void handleMessage(@NonNull TextMessage message) {
        String payload = message.getPayload();
        if (payload.equals("Connected to the ScoreSaber WSS")) { // Ignore the connection message
            return;
        }

        ScoreSaberWebsocketDataToken response = this.objectMapper.readValue(payload, ScoreSaberWebsocketDataToken.class);
        if (!response.getCommandName().equals("score")) { // Ignore non-score messages
            return;
        }

        // Decode the message using Jackson
        ScoreSaberPlayerScoreToken scoreToken = this.objectMapper.readValue(response.getCommandData().toString(), ScoreSaberPlayerScoreToken.class);
        ScoreSaberScoreToken score = scoreToken.getScore();
        ScoreSaberLeaderboardToken leaderboard = scoreToken.getLeaderboard();
        ScoreSaberScoreToken.LeaderboardPlayerInfo player = score.getLeaderboardPlayerInfo();

        if (!userService.validateSteamId(player.getId())) { // Validate the Steam ID
            return;
        }

        double accuracy = ((double) score.getBaseScore() / leaderboard.getMaxScore()) * 100;
        String difficulty = ScoreSaberUtils.parseDifficulty(leaderboard.getDifficulty().getDifficulty());
        double pp = platformService.getScoreSaberPlatform().getPp(leaderboard.getStars(), accuracy); // Recalculate the PP

        try (Sender sender = questDBService.getSender()) {
            sender.table("score")
                    .symbol("platform", Platform.Platforms.SCORESABER.getPlatformName())
                    // Player information
                    .symbol("player_id", player.getId())
                    // Score information
                    .symbol("leaderboard_id", leaderboard.getId())
                    .symbol("score_id", score.getId())
                    .doubleColumn("pp", pp)
                    .longColumn("rank", score.getRank())
                    .longColumn("score", score.getBaseScore())
                    .longColumn("modified_score", score.getModifiedScore())
                    .doubleColumn("weight", score.getWeight())
                    .stringColumn("modifiers", score.getModifiers())
                    .doubleColumn("multiplier", score.getMultiplier())
                    .longColumn("missed_notes", score.getMissedNotes())
                    .longColumn("bad_cuts", score.getBadCuts())
                    .longColumn("max_combo", score.getMaxCombo())
                    .doubleColumn("accuracy", accuracy)
                    .stringColumn("difficulty", difficulty)
                    .atNow();
        }
        log.info("Tracked score for {} with a score of {} and {}pp on {} with a rank of {}",
                player.getId(), score.getBaseScore(), score.getPp(), leaderboard.getId(), score.getRank());
    }
}
